//----------------------------------------------------------------------------
//
// File:        bitmap.cpp
// Date:        06-Apr-2000
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 2000-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <memory.h>
#include <stdio.h>
#include "common.hpp"
#include "logger.hpp"
#include "tms9900.hpp"
#include "tms9918a.hpp"
#include "ti994a.hpp"
#include "SDL.h"
#include "bitmap.hpp"
#include "tms9918a-sdl.hpp"

DBG_REGISTER ( __FILE__ );

cBitMap::cBitMap ( SDL_Surface *surface ) :
    m_Width ( 0 ),
    m_Height ( 0 ),
    m_Pitch ( 0 ),
    m_pSurface ( surface )
{
    FUNCTION_ENTRY ( this, "cBitMap ctor", true );

    if ( m_pSurface != NULL ) {
      m_Width  = m_pSurface->w;
      m_Height = m_pSurface->h;
      m_Pitch  = m_pSurface->pitch;
    }
}

cBitMap::~cBitMap ()
{
    FUNCTION_ENTRY ( this, "cBitMap dtor", true );

    SDL_FreeSurface ( m_pSurface );
}

void cBitMap::SetPalette ( SDL_Color *palette, int count )
{
    FUNCTION_ENTRY ( this, "cBitMap::SetPalette", true );

    SDL_SetColors ( m_pSurface, palette, 0, count );
}

void cBitMap::LockSurface ()
{
    FUNCTION_ENTRY ( this, "cBitMap::LockSurface", true );

    SDL_LockSurface ( m_pSurface );
}

void cBitMap::UnlockSurface ()
{
    FUNCTION_ENTRY ( this, "cBitMap::UnlockSurface", true );

    SDL_UnlockSurface ( m_pSurface );
}

void cBitMap::Copy ( cBitMap *original )
{
    FUNCTION_ENTRY ( this, "cBitMap::Copy", true );

    int scale = Width () / original->Width ();

    // See if we can use the SDL blit
    if ( scale == 1 ) {

        SDL_Rect srcRect = { 0, 0, original->Width (), original->Height () };
        SDL_Rect dstRect = { 0, 0, m_Width, m_Height };

        int wDif = Width () - original->Width ();
        int hDif = Height () - original->Height ();

        if ( wDif != 0 ) {
            if ( wDif > 0 ) {
                dstRect.x = wDif / 2;
            } else {
                srcRect.x = -wDif / 2;
                srcRect.w = Width ();
            }
        }

        if ( hDif != 0 ) {
            if ( hDif > 0 ) {
                dstRect.y = hDif / 2;
            } else {
                srcRect.y = -hDif / 2;
                srcRect.h = Height ();
            }
        }

        SDL_BlitSurface ( original->m_pSurface, &srcRect, m_pSurface, &dstRect );

    } else {

        UCHAR *pSrcData = original->GetData ();
        UCHAR *pDstData = GetData ();

        LockSurface ();
        original->LockSurface ();

        // Special purpose 2x scaling code
        int width  = original->Width ();
        int height = original->Height ();

        switch ( m_pSurface->format->BytesPerPixel ) {
            case 1 :
                {
                    for ( int y = 0; y < height; y++ ) {
                        for ( int x = 0; x < width; x++ ) {
                            pDstData [2*x] = pDstData [2*x+1] = pSrcData [x];
                        }
                        memcpy ( pDstData + Pitch (), pDstData, Width ());
                        pSrcData += original->Pitch ();
                        pDstData += 2 * Pitch ();
                    }
                }
                break;
            case 2 :
                {
                    for ( int y = 0; y < height; y++ ) {
                        for ( int x = 0; x < width; x++ ) {
                            (( USHORT * ) pDstData) [2*x] = (( USHORT * ) pDstData) [2*x+1] = (( USHORT * ) pSrcData) [x];
                        }
                        memcpy ( pDstData + Pitch (), pDstData, Width () * 2 );
                        pSrcData += original->Pitch ();
                        pDstData += 2 * Pitch ();
                    }
                }
                break;
            case 4 :
                {
                    for ( int y = 0; y < height; y++ ) {
                        ULONG *pData = ( ULONG * ) pDstData;
                        for ( int x = 0; x < width; x++ ) {
                            pData [0] = pData [1] = (( ULONG * ) pSrcData) [x];
                            pData += 2;
                        }
                        memcpy ( pDstData + Pitch (), pDstData, Width () * 4 );
                        pSrcData += original->Pitch ();
                        pDstData += 2 * Pitch ();
                    }
                }
                break;
        }

        original->UnlockSurface ();
        UnlockSurface ();
    }
}
