//----------------------------------------------------------------------------
//
// File:        disk.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description: A simple program to display the contents of a TI-Disk image
//              created by AnaDisk.  It also supports extracting files from
//              the disk and optionally converting them to 'standard' PC type
//              files.
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"

typedef unsigned char  BYTE;
typedef unsigned char  UCHAR;
typedef unsigned short USHORT;
typedef unsigned long  ULONG;

#define MAX_SECTORS     18
#define MAX_TRACKS      80

#if defined ( LITTLE_ENDIAN )
    #define SWAP_ENDIAN_16(x)   (((( x ) & 0xFF ) << 8 ) | ((( x ) >> 8 ) & 0xFF ))
#else
    #define SWAP_ENDIAN_16(x)   x
#endif

struct sSectorInfo {
    BYTE   ActualCylinder;
    BYTE   ActualSide;
    BYTE   LogicalCylinder;
    BYTE   LogicalSide;
    BYTE   LogicalSector;
    BYTE   Length;
    USHORT DataCount;
};

struct VIB {
    char    VolumeName [ 10 ];
    USHORT  FormattedSectors;
    UCHAR   SectorsPerTrack;
    char    DSK [3];
    UCHAR   reserved;
    UCHAR   TracksPerSide;
    UCHAR   Sides;
    UCHAR   Density;
    UCHAR   reserved2 [36];
    UCHAR   AllocationMap [200];
};

// Flags used by TI to indicate file types

#define DATA_TYPE		0x00
#define PROGRAM_TYPE		0x01

#define DISPLAY_TYPE		0x00
#define INTERNAL_TYPE		0x02

#define WRITE_PROTECTED_TYPE	0x08

#define FIXED_TYPE		0x00
#define VARIABLE_TYPE		0x80

struct CHAIN {
    UCHAR  start;
    UCHAR  start_offset;
    UCHAR  offset;
};

// On-Disk structure used by TI

struct sFileDescriptorRecord {
    char    FileName [ 10 ];
    char    reserved1 [ 2 ];
    UCHAR   FileStatus;
    UCHAR   RecordsPerSector;
    USHORT  TotalSectors;
    UCHAR   EOF_Offset;
    UCHAR   RecordLength;
    USHORT  NoFixedRecords;
    char    reserved2 [ 8 ];
    CHAIN   DataChain [ 76 ];
};

//----------------------------------------------------------------------------
//     +------+------+------+------+------+------+----------+
//     | ACYL | ASID | LCYL | LSID | LSEC | LLEN |  COUNT   |
//     +------+------+------+------+------+------+----------+
//
//      ACYL    Actual cylinder, 1 byte
//      ASID    Actual side, 1 byte
//      LCYL    Logical cylinder; cylinder as read, 1 byte
//      LSID    Logical side; or side as read, 1 byte
//      LSEC    Sector number as read, 1 byte
//      LLEN    Length code as read, 1 byte
//      COUNT   Byte count of data to follow, 2 bytes.  If zero,
//              no data is contained in this sector.
//
// All sectors occurring on a side will be grouped together;
// however, they will appear in the same order as they occurred on
// the diskette.  Therefore, if an 8 sector-per-track diskette were
// scanned which had a physical interleave of 2:1, the sectors might
// appear in the order 1,5,2,6,3,7,4,8 in the DOS dump file.
//
// After the last specified cylinder has been written to the DOS
// file, AnaDisk returns to the Main Menu.
//
//----------------------------------------------------------------------------

struct sSector {
    sSectorInfo  Header;
    BYTE        *Data;
};

struct sTrack {
    sSector      Sector [ MAX_SECTORS ];
};

struct sDiskInfo {
    VIB         *vib;
    int          noTracks;
    int          noSectorsPerTrack;
    sTrack       Track [ MAX_TRACKS ];
};

void ConvertFilename ( char *buffer )
{
    char *end = buffer;
    while ( *end ) end++;
    while ( *--end == ' ' );
    *++end = '\0';
    for ( ; *buffer; buffer++ ) {
        if ( isalnum ( *buffer )) continue;
        if ( *buffer == '_' ) continue;
        if ( *buffer == '-' ) continue;
        *buffer = '_';
    }
}

sSector *FindSector ( sDiskInfo *disk, int index )
{
    int t = index / disk->noSectorsPerTrack;
    int s = index % disk->noSectorsPerTrack;

    if ( t >= MAX_TRACKS ) {
        fprintf ( stderr, "Invalid sector index (%d)\n", index );
        exit ( - 1 );
        return NULL;
    }

    return &disk->Track[t].Sector[s];
}

bool ReadDisk ( sDiskInfo *disk, char *filename )
{
    FILE *diskFile = fopen ( filename, "rb" );
    if ( diskFile == NULL ) {
        fprintf ( stderr, "Unable to open file %s.\n", filename );
        return false;
    }

    memset ( disk, 0, sizeof ( sDiskInfo ));

    disk->noSectorsPerTrack = 0;

    int track = 0, sector = 0;
    while ( ! feof ( diskFile )) {
        sSectorInfo info;
        if ( fread ( &info, sizeof ( sSectorInfo ), 1, diskFile ) != 1 ) break;
        BYTE *ptr = info.DataCount ? new BYTE [ info.DataCount ] : NULL;
        if ( ptr ) fread ( ptr, info.DataCount, 1, diskFile );
        if ( info.ActualCylinder != track ) {
            if ( sector > disk->noSectorsPerTrack ) {
                disk->noSectorsPerTrack = sector;
            }
            track += 1;
            sector = 0;
        }
        if (( sector >= MAX_SECTORS ) || ( track >= MAX_TRACKS )) {
            fprintf ( stderr, "Error reading disk file\n" );
            return false;
        }
        disk->Track[track].Sector[sector].Header = info;
        disk->Track[track].Sector[sector++].Data = ptr;
    }

    fclose ( diskFile );

    disk->vib = ( VIB * ) disk->Track[0].Sector[0].Data;

    disk->noTracks = track;

    // If the disk has sectors/track, use it, otherwise we'll use our value
    if ( disk->vib->SectorsPerTrack != 0 ) {
        if ( disk->vib->SectorsPerTrack > MAX_SECTORS ) {
            fprintf ( stderr, "Invalid sector/track in VIB\n" );
            exit ( -1 );
        }
        disk->noSectorsPerTrack = disk->vib->SectorsPerTrack;
    }

    return true;
}

void DumpSectors ( sDiskInfo *disk, int index, int count, int isLast, sFileDescriptorRecord *FDR, bool convertFiles, FILE *outFile )
{
    for ( int i = 0; i < count; i++ ) {
        sSector *sector = FindSector ( disk, index );
        BYTE *ptr = ( BYTE * ) sector->Data;
        if ( FDR->FileStatus & PROGRAM_TYPE ) {
            int size = (( i == count - 1 ) && isLast ) ? FDR->EOF_Offset : 256;
            fwrite ( ptr, size, 1, outFile );
        } else if ( FDR->FileStatus & VARIABLE_TYPE ) {
            while ( *ptr != 0xFF ) {
                if (( convertFiles == true ) && (( FDR->FileStatus & INTERNAL_TYPE ) == 0 )) {
                    int length = *ptr++;
                    fwrite ( ptr, length, 1, outFile );
                    fputs ( "\r\n", outFile );
                    ptr += length;
                } else {
                    int length = *ptr + 1;
                    fwrite ( ptr, length, 1, outFile );
                    ptr += length;
                }
            }
        } else {
            for ( int j = 0; j < FDR->RecordsPerSector; j++ ) {
                fwrite ( ptr, FDR->RecordLength, 1, outFile );
                if ( convertFiles && ( FDR->FileStatus & INTERNAL_TYPE ) == 0 ) {
                    fputs ( "\r\n", outFile );
                }
                ptr += FDR->RecordLength;
            }
        }
        index++;
    }
}

void DumpFile ( sDiskInfo *disk, char *name, bool convertFiles )
{
    USHORT *dirIndex = ( USHORT * ) disk->Track[0].Sector[1].Data;
    while ( *dirIndex ) {
        int index = SWAP_ENDIAN_16 ( *dirIndex );
        sSector *sector = FindSector ( disk, index );
        sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) sector->Data;
        if ( strnicmp ( FDR->FileName, name, strlen ( name )) == 0 ) {
            // Find a unique filename to use
            char buffer [ 80 ];
            FILE *outFile;
            for ( int i = 0; i < 255; i++ ) {
                sprintf ( buffer, i ? "%s.%03d" : "%s", name, i );
                ConvertFilename ( buffer );
                outFile = fopen ( buffer, "rb" );
                if ( outFile == NULL ) break;
                fclose ( outFile );
            }

            // Open the file
            outFile = fopen ( buffer, "wb" );
            if ( outFile == NULL ) return;

            // Write the file header
            if (( convertFiles == false ) && ( FDR->FileStatus & INTERNAL_TYPE ) == 0 ) {
                UCHAR type, length;
                if ( FDR->FileStatus & PROGRAM_TYPE ) {
                    type   = 0x80;
                    length = 128;
                } else {
                    type   = ( FDR->FileStatus & INTERNAL_TYPE ) ? 0x08 : 0x00;
                    type  |= ( FDR->FileStatus & VARIABLE_TYPE ) ? 0x10 : 0x00;
                    length = FDR->RecordLength;
                }
                type |= ( FDR->FileStatus & WRITE_PROTECTED_TYPE ) ? 0x00 : 0x00;
                fwrite ( &type, 1, 1, outFile );
                fwrite ( &length, 1, 1, outFile );
            }

            int size = SWAP_ENDIAN_16 ( FDR->TotalSectors );
            int copied = 0;
            CHAIN *chain = FDR->DataChain;
            do {
                int sector = chain->start + (( int ) ( chain->start_offset & 0x0F ) << 8 );
                int offset = (( int ) chain->offset << 4 ) + ( chain->start_offset >> 4 ) + 1;
                DumpSectors ( disk, sector, offset - copied, offset == size, FDR, convertFiles, outFile );
                copied = offset;
                chain++;
            } while ( copied < size );
            fclose ( outFile );
            break;
        }
        dirIndex++;
    }
}

void ShowDirectory ( sDiskInfo *disk, bool dumpFiles, bool convertFiles )
{
    printf ( "Directory of %10.10s\n", disk->vib->VolumeName );
    printf ( "\n" );
    printf ( " Filename   Sectors   Type\n" );

    USHORT *dirIndex = ( USHORT * ) disk->Track[0].Sector[1].Data;
    while ( *dirIndex ) {
        int index = SWAP_ENDIAN_16 ( *dirIndex );
        sSector *sector = FindSector ( disk, index );
        sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) sector->Data;
        printf ( "  %10.10s", FDR->FileName );
        int size = SWAP_ENDIAN_16 ( FDR->TotalSectors );
        printf ( " %3d", size );
        if ( FDR->FileStatus & PROGRAM_TYPE ) {
	    printf ( "      PROGRAM    " );
        } else {
            printf ( "      %s/%s %3d", ( FDR->FileStatus & INTERNAL_TYPE ) ? "INT" : "DIS",
                                        ( FDR->FileStatus & VARIABLE_TYPE ) ? "VAR" : "FIX",
                                        FDR->RecordLength );
        }
        if ( dumpFiles == true ) {
            DumpFile ( disk, FDR->FileName, convertFiles );
        }

        printf ( "\n" );
        dirIndex++;
    }
}

int main ( int argc, char *argv[] )
{
    printf ( "TI-99/4A Diskette Viewer\n" );

    bool dumpFiles    = false;
    bool convertFiles = false;

    int index = 1;

    while ( index < argc ) {
        if ( argv [index][0] != '-' ) break;
        const char *ptr = &argv [index][1];
        while ( *ptr ) {
            switch ( toupper ( *ptr++ )) {
                case 'D' :
                    dumpFiles = true;
                    break;
                case 'C' :
                    convertFiles = true;
                    break;
                default :
                    fprintf ( stderr, "Unrecognized option %s\n", argv [index] );
                    break;
            }
        }
        index++;
    }

    if ( index >= argc ) {
        fprintf ( stderr, "No input file specified\n" );
        return -1;
    }

    sDiskInfo disk;

    if ( ReadDisk ( &disk, argv[index++] ) == false ) {
        return -1;
    }

    ShowDirectory ( &disk, dumpFiles, convertFiles );

    while ( index < argc ) {
        DumpFile ( &disk, argv[index++], convertFiles );
    }

    return 0;
}
