//----------------------------------------------------------------------------
//
// File:        mkspch.cpp
// Date:        06-May-2001
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"

struct sNode {
    const char *phrase;
    void       *data;
    int         length;
    sNode      *prev;
    sNode      *next;
    int         phrase_offset;
    int         data_offset;
};

UCHAR ROM [ 0x8000 ];

sNode *ReadNode ( FILE *file )
{
    static char null_data = 0xF0;
    static int line = 0;

    char buffer [512];
    char *ptr;

    // Read the next line from the file - ignore blank or comment lines
    do {
        line++;
        if ( fgets ( buffer, sizeof ( buffer ) - 1, file ) == 0 ) return NULL;
        buffer [511] = '\0';
        ptr = buffer;
        while ( isspace ( *ptr )) ptr++;
    } while (( *ptr == '#' ) || ( *ptr == '\0' ));

    // Remove the trailing '\n'
    buffer [strlen (buffer)-1] = '\0';

    char *end = ptr;
    while ( ! isspace ( *end )) end++;
    if ( end == ptr ) {
        fprintf ( stdout, "Invalid phrase on line %d\n", line );
        exit ( -1 );
    }

    if ( *end == '\0' ) {
        fprintf ( stdout, "Filename required on line %d\n", line );
        exit ( -1 );
    }

    *end = '\0';

    sNode *node = new sNode;
    memset ( node, 0, sizeof ( sNode ));

    node->phrase = strdup ( ptr );

    char *name = end + 1;
    while ( isspace ( *name )) name++;

    if ( strcmp ( name, "<null>" ) == 0 ) {
        node->data = &null_data;
        node->length = 1;
    } else {
        FILE *file = fopen ( name, "rb" );
        if ( file == NULL ) {
            fprintf ( stderr, "Invalid file '%s' specified on line %d\n", name, line );
            exit ( -1 );
        }
        fseek ( file, 0, SEEK_END );
        int size = ftell ( file );
        fseek ( file, 0, SEEK_SET );
        node->data   = new char [ size ];
        node->length = size;
        fread ( node->data, size, 1, file );
        fclose ( file );
    }

    return node;
}

int sort_node ( const void *ptr1, const void *ptr2 )
{
    const sNode *node1 = * ( const sNode ** ) ptr1;
    const sNode *node2 = * ( const sNode ** ) ptr2;

    return strcmp ( node1->phrase, node2->phrase );
}

sNode *SplitList ( sNode **list, int size )
{
    if ( size == 0 ) return NULL;

    int rootIndex = size / 2;
    sNode *root = list [ rootIndex ];

    root->prev = SplitList ( list, rootIndex );
    root->next = SplitList ( list + rootIndex + 1, size - rootIndex - 1 );

    return root;
}

void CalculateOffsets ( sNode *root, sNode **list, int count )
{
    // Calculate the phrase offsets
    int offset = 1;
    root->phrase_offset = offset;
    offset += 1 + strlen ( root->phrase ) + 2 + 2 + 4;
    for ( int i = 0; i < count; i++ ) {
        sNode *node = list [i];
        if ( node == root ) continue;
        node->phrase_offset = offset;
        offset += 1 + strlen ( node->phrase ) + 2 + 2 + 4;
    }

    // NOTE: The following is a simple-minded approach.  It would be better
    //   to look for an existing match before adding the new data pattern so
    //   that, for example, '1' and 'ONE' would point to the same memory.

    // Calculate the data offsets
    for ( int i = 0; i < count; i++ ) {
        sNode *node = list [i];
        node->data_offset = offset;
        offset += node->length;
    }
}

UCHAR *StoreNode ( sNode *node, UCHAR *ptr )
{
    *ptr++ = strlen ( node->phrase );

    strcpy (( char * ) ptr, node->phrase );
    ptr += ptr [-1];

    int offset;

    offset = node->prev ? node->prev->phrase_offset : 0;
    *ptr++ = offset >> 8;
    *ptr++ = offset & 0xFF;

    offset = node->next ? node->next->phrase_offset : 0;
    *ptr++ = offset >> 8;
    *ptr++ = offset & 0xFF;

    offset = node->data_offset;
    *ptr++ = '\0';              // Dunno what this is yet
    *ptr++ = offset >> 8;
    *ptr++ = offset & 0xFF;
    *ptr++ = '\0';              // Dunno what this is yet
    
    return ptr;
}

void MakeROM ( sNode *root, sNode **list, int count )
{
    UCHAR *ptr = ROM;

    *ptr++ = 0xAA;

    // Store the phrases
    ptr = StoreNode ( root, ptr );
    for ( int i = 0; i < count; i++ ) {
        sNode *node = list [i];
        if ( node == root ) continue;
        ptr = StoreNode ( node, ptr );
    }

    // Store the data
    for ( int i = 0; i < count; i++ ) {
        sNode *node = list [i];
        memcpy ( ptr, node->data, node->length );
        ptr += node->length;
    }
}

int main ( int, char *argv[] )
{
    FILE *file = fopen ( argv [1], "rt" );
    if ( file == NULL ) {
        fprintf ( stderr, "Unable to open file '%s'\n", argv [1]);
        return -1;
    }

    sNode *list [ 3277 ];
    int count = 0;

    // Read in the basic phrases and their speech data
    for ( EVER ) {
        sNode *node = ReadNode ( file );
        if ( node == NULL ) break;
        list [count++] = node;
    }

    fclose ( file );

    fprintf ( stdout, "%d phrases read\n", count );

    // Put them in alphabetical order
    qsort ( list, count, sizeof ( sNode * ), sort_node );

    // Build the binary tree
    sNode *root = SplitList ( list, count );

    CalculateOffsets ( root, list, count );

    MakeROM ( root, list, count );

    FILE *rom = fopen ( "spchrom.bin", "wb" );
    if ( rom == NULL ) {
        fprintf ( stderr, "Unable to open output file spchrom.dat\n" );
        return -1;
    }

    fwrite ( ROM, sizeof ( ROM ), 1, rom );

    fclose ( rom );

    return 0;
}
